import numpy as np
import matplotlib.pyplot as plt
import PIL as P

from .coords import Drawable, mk_domain
from .objects import TileSet


class Image:
    """
    Our base representation of image data, perfectly functional for the
    majority of use cases.

    pixels : ndarry
        The numpy array holding the raw image data
    """

    def __init__(self, width=None, height=None, domain=None,
                 background=(255, 255, 255, 255), xAA=1, pixels=None):
        """
        Image Constructor

        There are two ways to construct an image, the first is to specify the
        width and height in pixels plus a fill color.

        Or you can provide a numpy array with the shape (height, width, 4)
        which you can use as an image.

        It's important to note that if you provide the pixels argument it will
        override all the others.

        width : int, optional
            The width of the image in pixels. Default None
        height : int, optional. Default None
            The height of the image in pixels
        domain : (int, int) -> np.meshgrid-like
            This is a function which given a width and a height returns a grid
            containing the mathematical points associated with each pixel.
            Default: None
        background : 4-tuple, optional
            The fill RGBA colour of the image, each value can be in the range
            0-255 (default is (0, 0, 0, 0) - fully transparent black)
        pixels : numpy array, optional
            If you already have a numpy array you want to use as an image
            simply pass it in here. **Note:** The array must have the shape
            (height, width, 4). Default value None.
        """

        self._domain = domain
        self._xAA = xAA

        if pixels is not None:
            shape = pixels.shape

            if len(shape) != 3 or shape[-1] != 4:
                raise ValueError("Pixels array must have shape:"
                                 " (height, width, 4)")

            self.pixels = pixels
            self._width = shape[1]
            self._height = shape[0]
        else:
            if width is None or height is None:
                raise ValueError("If no pixels array is given you"
                                 " must specify a width and height")

            # If given an RGB lookalike, convert it to RGBA
            if len(background) == 3:
                background = tuple([*background, 255])

            self._height = height
            self._width = width

            self.pixels = np.full((height * xAA, width * xAA, 4), background,
                                  dtype=np.uint8)

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def color(self):
        return self.pixels[:, :, 0:3]

    @color.setter
    def color(self, value):
        self.pixels[:, :, 0:3] = value

    @property
    def red(self):
        return self.pixels[:, :, 0]

    @red.setter
    def red(self, value):
        self.pixels[:, :, 0] = value

    @property
    def green(self):
        return self.pixels[:, :, 1]

    @green.setter
    def green(self, value):
        self.pixels[:, :, 1] = value

    @property
    def blue(self):
        return self.pixels[:, :, 2]

    @blue.setter
    def blue(self, value):
        self.pixels[:, :, 2] = value

    @property
    def alpha(self):
        return self.pixels[:, :, 3]

    @alpha.setter
    def alpha(self, value):
        self.pixels[:, :, 3] = value

    @property
    def xAA(self):
        return self._xAA

    @property
    def domain(self):
        return self._domain

    @domain.setter
    def domain(self, value):
        self._domain = value

    @classmethod
    def fromarray(cls, px):
        return cls(pixels=px)

    @classmethod
    def fromfile(cls, filename):
        img = P.Image.open(filename)
        arr = np.array(img.getdata(), dtype=np.uint8)
        arr = arr.reshape(img.size[1], img.size[0], 4)
        return Image(pixels=arr)

    def _flip_index(self, index):
        """
        The underlying representation is in (y,x) 'coordinates' which
        breaks normal mental models of coordinates being (x,y). So in
        an effort to keep this mental model consistent throughout this
        method reorders a given index in that manner.

        If index is just a number, we assume it to be indexing the x
        coordinate so we return (:, x)

        If index is tuple of indices we swap the first two entries
        """

        if isinstance(index, (tuple,)):
            if len(index) == 2:
                return (index[1], index[0])
            else:
                return (index[1], index[0], *index[2:])
        else:
            return (slice(None), index)

    def _get_by_pixels(self, index):
        """
        Once we have decided that the index represents pixels
        this method is responsible for interpreting it and
        returning the appropriate object
        """
        px = self.pixels[index]
        shape = px.shape

        if len(shape) == 3:
            return Image.fromarray(px)
        else:
            return px

    def _is_pixel_addr(self, index):
        """
        This method is for deciding whether or not a given
        index represents a pixel index or mathematical points
        """

        for idx in index:

            # Is this item a simple integer - hence pixel address?
            if isinstance(idx, (int,)):
                continue

            # Is it a float?
            if isinstance(idx, (float,)):
                return False

            # Is it a slice?
            if isinstance(idx, (slice,)):

                if isinstance(idx.start, (float,)) or\
                   isinstance(idx.stop, (float,)):
                    return False

        # If we get this far, then it must be a pixel address
        return True

    def __getitem__(self, index):
        idx = self._flip_index(index)

        if self._is_pixel_addr(idx):
            return self._get_by_pixels(idx)

        # If we get this far, we must be dealing with a 'mathematical
        # address' So we 'simply' have to figure out which pixels are mapped
        # to the mathematial points in question

        # This involves using the built-in domain function - which
        # of course has to exist for this to work
        if self._domain is None:
            raise RuntimeError('In order to index pixels by mathematical '
                               'points the image must be associated with '
                               'a domain!')

        # For the moment we will restrict ourselves to the index being made
        # up of 2 slices. This might change eventually
        if not isinstance(idx[0], (slice,)) or\
           not isinstance(idx[1], (slice,)):
            raise ValueError('Mathematical indexing is currently only '
                             'supported for slices')

        # Step 0: Unpack the slices
        yslice = idx[0]
        ystart = yslice.start
        ystop = yslice.stop

        xslice = idx[1]
        xstart = xslice.start
        xstop = xslice.stop

        # Step 1: Evaluate the domain function to get the points
        (xs, *_), YS = self._domain(self.width * self.xAA,
                                    self.height * self.xAA)
        ys = np.array([it[0] for it in YS])

        # Step 2: See which pixels the values are closest to
        xstart_d = [(None, None)] if xstart is None else\
                   sorted([(abs(x - xstart), i) for i, x in enumerate(xs)])

        xstop_d = [(None, None)] if xstop is None else\
                  sorted([(abs(x - xstop), i) for i, x in enumerate(xs)])

        ystart_d = [(None, None)] if ystart is None else\
                   sorted([(abs(y - ystart), i) for i, y in enumerate(ys)])

        ystop_d = [(None, None)] if ystop is None else\
                  sorted([(abs(y - ystop), i) for i, y in enumerate(ys)])

        # Step 3: Construct new slice objects with the new pixel addresses
        # and pass it off to the _get_by_pixels method
        slice_x = slice(xstart_d[0][1], xstop_d[0][1], None)
        slice_y = slice(ystart_d[0][1], ystop_d[0][1], None)

        return self._get_by_pixels((slice_y, slice_x))

    def __setitem__(self, index, value):

        # If we are given something that looks like RGB
        # convert it to RGBA
        if len(value) == 3:
            value = tuple([*value, 255])

        self.pixels[self._flip_index(index)] = value

    def __repr__(self):
        shape = self.pixels.shape
        return '%ix%i Image' % (shape[1], shape[0])

    def __neg__(self):
        """
        Easily obtain the negative of an image by typing
        (-img)
        """

        # Create the function that will invert the values of
        # the color pixels and vectorize it
        neg = lambda v: abs(v - 255)
        vneg = np.vectorize(neg, otypes=(np.uint8,))

        # Create the new pixel array
        alphas = self.alpha
        alphas.shape = (self.height, self.width, 1)
        colors = vneg(self.color)

        px = np.append(colors, alphas, axis=2)

        return Image(pixels=px)

    def __and__(self, other):
        """
        Implementation of the boolean AND, best when used with black and
        white images
        """

        if not isinstance(other, (Image,)):
            raise TypeError('AND is only supported between instances of '
                            'the Image class')

        if self.pixels.shape != other.pixels.shape:
            raise ValueError('AND can only be used with Images that have '
                             'the same dimensions!')

        ax = self.pixels
        bx = other.pixels

        # Scale the pixels of b
        scaled = bx / 255

        # Do the multiplication - it's important to note that the entries
        # of cx are now 'float64'
        cx = ax * scaled

        # Create a new numpy array so that the conversion back to uint8 is
        # correct
        px = np.array(cx, dtype=np.uint8)

        return Image.fromarray(px)

    def __call__(self, f, overwrite_domain=True, use_host_domain=False):
        """
        Implementing this 'magic' method allows us to use img(drawable)
        to trigger the 'drawing' of an object onto an arbitrary image.

        The default behavior of this function is to take the drawable
        f, extract it's domain function and proceed to compute the mask
        and color of the affected pixels. Once that is done, the extracted
        domain function is stored internally so that users may optionally
        index the pixels by mathematical points, instead of raw pixels.

        However, it is possible to override this behavior with the two
        options:
            - use_host_domain -> False by default
            - overwrite_domain -> True by default

        If use_host_domain is True, instead of using the domain that
        comes with the drawable, the domain already associated with
        the underlying image is used instead.

        If overwrite_domain is False, the image keeps its original
        domain, instead of saving the new one that the drawable
        possesses
        """

        # Of course, the image has to have a domain already if we are
        # to use this one instead
        if use_host_domain and self._domain is None:
            raise RuntimeError('The image does not have an associated '
                               'domain!')

        # Get the domain from the correct place
        if use_host_domain:
            XS, YS = self._domain(self.width * self.xAA,
                                  self.height * self.xAA)
        else:
            XS, YS = f.domainfunc(self.width * self.xAA,
                                  self.height * self.xAA)

        # Compute the mask
        mask = f.maskfunc(XS, YS)

        # When it comes to the color function, there are some cases
        # to consider. If the 'function' is in fact just a tuple
        # then we don't have to evaluate it across the entire domain
        colorfunc = f.colorfunc

        if isinstance(colorfunc, (tuple,)):
            self.pixels[mask] = colorfunc
        else:
            self.pixels[mask] = colorfunc(XS[mask], YS[mask])

        # Only now that all is said and done do we overwrite the domain,
        # that way we don't get something halfway between one or the other
        if overwrite_domain and not use_host_domain:
            self._domain = f.domainfunc

    def show(self):
        """
        Show the image using a matplotlib figure - useful for use in
        notebooks
        """
        return plt.imshow(self.pixels)

    def save(self, filename):
        """
        Save the image to the given filename
        """

        image = P.Image.frombuffer('RGBA', (self.width * self.xAA,
                                   self.height * self.xAA), self.pixels,
                                   'raw', 'RGBA', 0, 1)

        # Apply our naive AA by downscaling the image
        if self.xAA > 1:
            image = image.resize((self.width, self.height), resample=P.Image.BICUBIC)

        with open(filename, 'wb') as f:
            image.save(f)


class LayeredImage:
    """
    This is a container for a number of images, with a number of conveniences
    for blending/merging them together - just as you would in GIMP/Photoshop
    etc.
    """

    def __init__(self, width=None, height=None, num_layers=4, imgs=None):

        if imgs is not None:

            # TODO: Do a check to make sure given images make sense
            self._layers = imgs
            self._width = imgs[0].width
            self._height = imgs[0].height

        else:
            self._layers = [Image(width, height, background=(0, 0, 0, 0))
                            for _ in range(num_layers)]

            self._width = width
            self._height = height

    def __repr__(self):
        return "%ix%i Layered Image with %i layers"\
                % (self.width, self.height, self.nlayers)

    def __getitem__(self, key):

        if isinstance(key, (int,)):
            return self._layers[key]

        else:
            return 'Other methods coming soon'

    @classmethod
    def fromlist(cls, imgs):
        return cls(imgs=imgs)

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def nlayers(self):
        return len(self._layers)


class TiledImage(Image):
    """
    This builds on the base image class, to allow easy creation of
    images from tilesets. Useful for constructing backgrounds based
    on a grid - e.g level designs etc.
    """

    def __init__(self, width, height, xmax=10, ymax=10,
                 xmin=0, ymin=0, *args, **kwargs):

        # First call the parent to set things up
        super().__init__(width, height, *args, **kwargs)

        # Now we will set up the domain for the image
        self._xmin = xmin
        self._xmax = xmax

        self._ymin = ymin
        self._ymax = ymax

        self.domain = mk_domain(xmin, xmax, ymin, ymax)

    def __call__(self, tileset, overwrite_domain=False):
        """
        This allows for the possibility of easy application
        of a tileset to the Image, but bog standard drawables
        are still ok.
        """

        # Are we dealing with a drawable?
        if isinstance(tileset, (Drawable,)):

            # If so do the "normal" thing
            super().__call__(tileset, overwrite_domain)
            return

        if isinstance(tileset, (TileSet,)):

            # Here we assume that the tileset is not time dependent
            # so just call the layout method with some, argument
            layout = tileset.layout(0, 0)
            tiles = tileset._tiles

        elif isinstance(tileset, (tuple,)):

            # Here we assume that we have been given a time dependent
            # tileset
            tiles, layout = tileset

        else:
            raise TypeError('TiledImages only support Drawables and '
                            'TileSets!!')

        # With that out of the way let's get down to it. First let's check
        # that the given tileset makes sense

        # Numpy freaks out if we try to .unique() an array containing None's
        # So just turn them into zeros, as a tileset will always have a
        # zeroth tile
        tozero = lambda v: v if v is not None else 0
        vtoz = np.vectorize(tozero)
        test_layout = vtoz(layout)

        if len(tiles) < len(np.unique(test_layout)):
            raise ValueError('The layout defined references more tiles '
                             'than are defined!')

        # Transpose the layout array so that it makes sense to the user
        layout = layout.transpose()

        # Next we need to decide on the size of each of the grid squares
        nx, ny = layout.shape
        xsize = (self.xmax - self.xmin) / nx
        ysize = (self.ymax - self.ymin) / ny

        # Now we go through each tile in the layout and apply it to the
        # image
        for j in range(ny):

            ystart = self.ymin + j*ysize
            yend = ystart + ysize

            for i in range(nx):
                xstart = self.xmin + i*xsize
                xend = xstart + xsize

                tileidx = layout[i, j]

                if tileidx is not None:
                    tile = tiles[tileidx]
                    self[xstart:xend, ystart:yend](tile)

    @property
    def xmin(self):
        return self._xmin

    @xmin.setter
    def xmin(self, value):

        if not isinstance(value, (int, float)):
            raise TypeError('xmin must be a number!')

        if value >= self.xmax:
            raise ValueError('xmin cannot be larger than xmax!')

        self._xmin = value
        self.domain = mk_domain(value, self.xmax, self.ymin, self.ymax)

    @property
    def xmax(self):
        return self._xmax

    @xmax.setter
    def xmax(self, value):

        if not isinstance(value, (int, float)):
            raise TypeError('xmax must be a number!')

        if value <= self.xmin:
            raise ValueError('xmax cannot be less than xmin!')

        self._xmin = value
        self.domain = mk_domain(self.xmin, value, self.ymin, self.ymax)

    @property
    def ymin(self):
        return self._ymin

    @ymin.setter
    def ymin(self, value):

        if not isinstance(value, (int, float)):
            raise TypeError('ymax must be a number!')

        if value >= self.ymax:
            raise ValueError('ymin cannot be larger than ymax!')

        self._ymin = value
        self.domain = mk_domain(self.xmin, self.xmax, value, self.ymax)

    @property
    def ymax(self):
        return self._ymax

    @ymax.setter
    def ymax(self, value):

        if not isinstance(value, (int, float)):
            raise TypeError('ymax must be a number!')

        if value <= self.ymin:
            raise ValueError('ymax cannot be less than ymin!')

        self._ymax = value
        self.domain = mk_domain(self.xmin, self.xmax, self.ymin, value)



import pytest
from hypothesis import given
from hypothesis.strategies import integers, tuples


import numpy as np
import numpy.random as npr
from stylo.image import Image


dimension = integers(min_value=2, max_value=5000)
smalldim = integers(min_value=2, max_value=256)
colvalue = integers(min_value=0, max_value=255)
color = tuples(colvalue, colvalue, colvalue, colvalue)


@given(width=smalldim, height=smalldim)
def test_width_height_init(width, height):

    img = Image(width, height)

    assert img.pixels.shape == (height, width, 4)
    assert img.pixels.dtype == np.uint8

    unique = np.unique(img.pixels)
    assert unique.shape == (1,)
    assert (unique == [(255, 255, 255, 255)]).all()


@given(width=smalldim, height=smalldim, background=color)
def test_width_height_background_init(width, height, background):

    img = Image(width, height, background=background)

    assert img.pixels.shape == (height, width, 4)
    assert img.pixels.dtype == np.uint8

    unique = np.unique(img.pixels, axis=0)
    unique = np.unique(unique, axis=1).flatten()
    assert unique.shape == (4,)
    assert (unique == [background]).all()


@given(width=smalldim, height=smalldim)
def test_pixel_init(width, height):

    px = npr.randint(0, 255, (height, width, 4), dtype=np.uint8)
    img = Image(pixels=px)

    assert img.pixels.shape == (height, width, 4)
    assert (img.pixels == px).all()

    img = Image.fromarray(px)

    assert img.pixels.shape == (height, width, 4)
    assert (img.pixels == px).all()


def test_init_exceptions():

    with pytest.raises(ValueError) as err:
        img = Image()

    assert 'specify a width and height' in str(err.value)

    with pytest.raises(ValueError) as err:
        img = Image(width=12)

    assert 'specify a width and height' in str(err.value)

    with pytest.raises(ValueError) as err:
        img = Image(height=12)

    assert 'specify a width and height' in str(err.value)


@given(width=smalldim, height=smalldim)
def test_pixel_init_exceptions(width, height):

    px = npr.randint(0, 255, (height, width), dtype=np.uint8)

    with pytest.raises(ValueError) as err:
        img = Image(pixels=px)

    assert 'must have shape:' in str(err.value)

    px = npr.randint(0, 255, (height, width, 5), dtype=np.uint8)

    with pytest.raises(ValueError) as err:
        img = Image(pixels=px)

    assert 'must have shape:' in str(err.value)


@given(width=smalldim, height=smalldim)
def test_repr(width, height):

    img = Image(width, height)
    assert repr(img) == '%ix%i Image' % (width, height)


@given(index=integers(min_value=4, max_value=512))
def test_getitem_pixels(index):

    img = Image(512, 512)

    # Test that slicing by x works as expected
    sliced_x = img[:index]

    assert isinstance(sliced_x, (Image,))
    assert sliced_x.width == index
    assert sliced_x.height == 512

    # Test that slicing by y works as expected
    sliced_y = img[:, :index]

    assert isinstance(sliced_y, (Image,))
    assert sliced_y.width == 512
    assert sliced_y.height == index

    # Test that slicing by both works as expected
    sliced_both = img[:index, :index]

    assert isinstance(sliced_both, (Image,))
    assert sliced_both.width == index
    assert sliced_both.height == index

    # As a further test ensure that the smaller images
    # point back to subsets of the original pixel
    # data
    sliced_both[:, :] = (255, 0, 0, 255)

    assert (img.pixels[:index, :index] == sliced_both.pixels).all()
    assert (img.pixels[index:, index:] == (255, 255, 255, 255)).all()



import unittest.mock as mock
import numpy as np


import pytest
from pytest import raises, fixture
from hypothesis import given, assume
from hypothesis.strategies import composite, integers,\
                                  tuples, lists, floats,\
                                  text


from stylo.interpolate import Channel

# Some settings
DATA_MAX = 1.0
DATA_MIN = 0.0
DATA_LENGTH = 512

# Some strategy definitions
index = integers(min_value=0, max_value=(DATA_LENGTH - 1))
pveint = integers(min_value=1, max_value=2*DATA_LENGTH)
real = floats(min_value=-1e6, max_value=1e6)


@composite
def data(draw):
    length = draw(pveint)
    start = draw(real)
    stop = draw(real)

    return np.linspace(start, stop, length)


segment = tuples(pveint, data())


# Some fixtures
@fixture
def channel():
    # Just a 'Vanilla' Channel with no extras
    return Channel()


@fixture
def channel_cycle():
    # An empty channel with Cycle enabled
    c = Channel()
    c._cycle = True

    return c


@fixture
def channel_data():
    # A channel with some example data
    c = Channel()
    c._segments = 1     # Pretend this data is real
    c._data = np.linspace(DATA_MIN, DATA_MAX, DATA_LENGTH)

    return c


@fixture
def channel_data_cycle():
    # A channel with cycled example data
    c = Channel()
    c._cycle = True
    c._segments = 1     # Pretend that the data is real
    c._data = np.linspace(DATA_MIN, DATA_MAX, DATA_LENGTH)

    return c


class TestProperties(object):

    def test_data_property(self):

        channel = Channel()

        # By default it should be empty
        assert (channel.data == np.array([])).all()

        # And it should be READ ONLY!
        with raises(AttributeError) as err:
            channel.data = 1

        assert 'can\'t set attribute' in str(err.value)


    def test_cycle_property(self):

        channel = Channel()

        # By default it should be false
        assert channel.cycle is False

        # We should be able to set it
        channel.cycle = True
        assert channel.cycle is True

        # If we try to set it to anything except a bool
        # then it should complain
        with raises(TypeError) as err:
            channel.cycle = 42

        assert 'must be a bool' in str(err.value)

        # We should also check that the old value was preserved
        assert channel.cycle is True

    @given(name=text(average_size=32))
    def test_name_property(self, name):

        channel = Channel()

        # By default this should be the empty string
        assert channel.name is ''

        # We should be able to change that
        channel.name = name
        assert channel.name == name

        # But only to a string
        with raises(TypeError) as err:
            channel.name = True

        assert 'must be a str' in str(err.value)

        # We should also check that the old value was preserved
        assert channel.name == name

    @given(fps=pveint)
    @mock.patch.object(Channel, '_construct')
    def test_fps_property_with_int(self, mock_construct, fps):

        channel = Channel()

        # The construct method should already have been called
        assert mock_construct.call_count == 1

        # By default this should be 25
        assert channel.FPS == 25

        # We should be able to change this
        channel.FPS = fps
        assert channel.FPS == fps

        # But changing this should call the _construct method
        assert mock_construct.call_count == 2

    @given(fps=pveint)
    @mock.patch.object(Channel, '_construct')
    def test_fps_property_with_bad_values(self, mock_construct, fps):

        channel = Channel()

        # FPS should be a default value of 25
        assert channel.FPS == 25

        # The construct method should have already been called
        assert mock_construct.call_count == 1

        # Try and change the FPS property to something
        # other than an int
        with raises(TypeError) as err:
            channel.FPS = 'a string!'

        assert 'must be an int' in str(err.value)

        # The value should be unchanged
        assert channel.FPS == 25

        # The construct method should NOT have been called again
        assert mock_construct.call_count == 1

        # Similarly there should be an error if we try a negative
        # FPS
        with raises(ValueError) as err:
            channel.FPS = -fps

        assert 'must be larger than 0' in str(err.value)
        assert channel.FPS == 25
        assert mock_construct.call_count == 1

    @given(segments=lists(segment, average_size=12))
    @mock.patch.object(Channel, '_construct')
    def test_segments_property(self, mock_construct, segments):

        channel = Channel()

        # By default it should be the empty list and the construct
        # method should've been called
        assert channel.segments == []
        assert mock_construct.call_count == 1

        # We should be able to change this
        channel.segments = segments

        mask = channel.segments == sorted(segments, key=lambda x: x[0])

        if isinstance(mask, (bool,)):
            assert mask
        else:
            assert all(mask)

        # And the construct function should be called again
        assert mock_construct.call_count == 2

    @mock.patch.object(Channel, '_construct')
    def test_segments_property_with_bad_values(self, mock_construct):

        channel = Channel()
        assert mock_construct.call_count == 1

        # The segments property must enforce the correct
        # structure
        with raises(TypeError) as err:
            channel.segments = 128

        assert 'must be a list' in str(err.value)

        # The old value must be unchanged and the construct
        # function not called
        assert channel.segments == []
        assert mock_construct.call_count == 1

        # Each item must be a tuple
        with raises(TypeError) as err:
            channel.segments = [(0, [1, 2, 3]), 1.2]

        assert 'must be a tuple' in str(err.value)
        assert channel.segments == []
        assert mock_construct.call_count == 1

        # Each tuple must be the right length
        with raises(TypeError) as err:
            channel.segments = [(0, [1, 2, 3]), (0.2,)]

        assert 'length 2' in str(err.value)
        assert channel.segments == []
        assert mock_construct.call_count == 1

        # The first entry must be a non negative integer
        with raises(TypeError) as err:
            channel.segments = [(0, [1, 2, 3]), (0.2, [])]

        assert 'must be an integer' in str(err.value)
        assert channel.segments == []
        assert mock_construct.call_count == 1

        with raises(ValueError) as err:
            channel.segments = [(0, [1, 2, 3]), (-2, [])]

        assert 'cannot be negative' in str(err.value)
        assert channel.segments == []
        assert mock_construct.call_count == 1

        # Finally the second entry must be an iterable
        with raises(TypeError) as err:
            channel.segments = [(0, [1, 2, 3]), (12, 128)]

        assert 'must be iterable' in str(err.value)
        assert channel.segments == []
        assert mock_construct.call_count == 1


class TestGetItem(object):

    @given(i=index)
    def test_with_index_nodata_nocycle(self, channel, i):

        # When a channel has no segments it should generate fake
        # data to make it look like it is zero everywhere
        assert channel[i] == 0.0
        assert channel[-i] == 0.0

    @given(i=index)
    def test_with_index_nodata_cycle(self, channel_cycle, i):

        # In this case cycle being set should not make a difference
        assert channel_cycle[i] == 0.0
        assert channel_cycle[-i] == 0.0

    @given(i=index, offset=pveint)
    def test_with_slice_nodata_nocycle(self, channel, i, offset):
        # When a channel has no segments it should generate
        # fake data to make it look like it is zero everywhere
        res = channel[slice(i, i + offset)]
        assert len(res) == offset
        assert (res == 0.0).all()

    @given(i=index, offset=pveint)
    def test_with_slice_nodata_cycle(self, channel_cycle, i, offset):

        # In this case cycle being set should not make a difference
        res = channel_cycle[slice(i, i + offset)]
        assert len(res) == offset
        assert (res == 0.0).all()

    @given(i=index)
    def test_with_inner_index_data_nocycle(self, channel_data, i):

        # An index within the data should return the data itself
        assert channel_data[i] == \
               np.linspace(DATA_MIN, DATA_MAX, DATA_LENGTH)[i]

    @given(i=index)
    def test_with_later_index_data_nocycle(self, channel_data, i):

        # If we're not cycling the data then any index after the data
        # should return the final value
        assert channel_data[DATA_LENGTH + i] == DATA_MAX

    @given(i=index)
    def test_with_earlier_index_data_nocycle(self, channel_data, i):
        # Negative indices should return the initial value of the data
        assert channel_data[-i] == DATA_MIN

    @given(i=index)
    def test_with_inner_index_data_cycle(self, channel_data_cycle, i):

        # An index within the data should return the data itself
        assert channel_data_cycle[i] == \
               np.linspace(DATA_MIN, DATA_MAX, DATA_LENGTH)[i]

    @given(i=index)
    def test_with_later_index_data_cycle(self, channel_data_cycle, i):
        # An index after the data should return the value of i mod length
        assert channel_data_cycle[DATA_LENGTH + i] == \
               np.linspace(DATA_MIN, DATA_MAX, DATA_LENGTH)[i]

    @given(i=index)
    def test_with_earlier_index_data_cycle(self, channel_data_cycle, i):
        assume(i > 0)

        # The same for any negative indices
        assert channel_data_cycle[-i] == \
               np.linspace(DATA_MIN, DATA_MAX,
                           DATA_LENGTH)[DATA_LENGTH - i]

    @given(i=index, offset=pveint)
    def test_with_inner_bounded_slice_data_nocycle(self, channel_data,
                                                   i, offset):
        assume(i + offset < DATA_LENGTH - 1)

        # Slices within the data should return the data itself
        res = channel_data[slice(i, i + offset)]
        assert len(res) == offset

        mask = res == np.linspace(DATA_MIN, DATA_MAX,
                                  DATA_LENGTH)[slice(i, i + offset)]
        assert mask.all()

    @given(i=index)
    def test_with_inner_boundbelow_slice_data_nocycle(self, channel_data, i):

        # Slices that are unbounded should stop at the end of the real data
        res = channel_data[slice(i, None)]
        assert len(res) == DATA_LENGTH - i

        mask = res == np.linspace(DATA_MIN, DATA_MAX,
                                  DATA_LENGTH)[slice(i, None)]
        assert mask.all()

    @given(i=index)
    def test_with_inner_boundedabove_slice_data_nocycle(self, channel_data, i):

        # Slices that are unbounded should simply start at the beginning
        # of the real data
        res = channel_data[slice(None, i)]
        assert len(res) == i

        mask = res == np.linspace(DATA_MIN, DATA_MAX,
                                  DATA_LENGTH)[slice(None, i)]
        assert mask.all()

    def test_with_unbounded_slice_data_nocycle(self, channel_data):

        # Unbounded slice should simply return the data
        res = channel_data[slice(None, None)]
        assert len(res) == DATA_LENGTH

        mask = res == np.linspace(DATA_MIN, DATA_MAX, DATA_LENGTH)
        assert mask.all()

    @given(offset=pveint)
    def test_with_later_slice_data_nocycle(self, channel_data, offset):

        # When the slice extends over the end of the data (with no cycling)
        # it should return the final value repeated
        res = channel_data[slice(DATA_LENGTH, DATA_LENGTH + offset)]
        assert len(res) == offset
        assert (res == 1.0).all()

    @given(offset=pveint)
    def test_with_earlier_slice_data_nocycle(self, channel_data, offset):

        # Similarly if the slice start before the data it should
        # return the initial value repeated
        res = channel_data[slice(0 - offset, 0)]
        assert len(res) == offset
        assert (res == 0.0).all()

    @given(i=index, offset=pveint)
    def test_with_inner_bounded_slice_data_cycle(self, channel_data_cycle,
                                                 i, offset):
        assume(i + offset < DATA_LENGTH)

        # When a channel has segments, a slice fully within the
        # data should return the data itself
        res = channel_data_cycle[slice(i, i + offset)]
        assert len(res) == offset

        mask = res == np.linspace(DATA_MIN, DATA_MAX,
                                  DATA_LENGTH)[slice(i, i + offset)]
        assert mask.all()

    @given(i=index)
    def test_with_inner_boundbelow_slice_data_cycle(self,
                                                    channel_data_cycle, i):

        # Unbounded slices should stop at the end of the real data
        res = channel_data_cycle[slice(i, None)]
        assert len(res) == DATA_LENGTH - i

        mask = res == np.linspace(DATA_MIN, DATA_MAX,
                                  DATA_LENGTH)[slice(i, None)]
        assert mask.all()

    @given(i=index)
    def test_with_inner_boundabove_slice_data_cycle(self,
                                                    channel_data_cycle, i):

        # Unbounded slices should just start at the beginning of the data
        res = channel_data_cycle[slice(None, i)]
        assert len(res) == i

        mask = res == np.linspace(DATA_MIN, DATA_MAX,
                                  DATA_LENGTH)[slice(None, i)]
        assert mask.all()

    @given(offset=pveint)
    def test_with_later_slice_data_cycle(self, channel_data_cycle, offset):

        # When the slice extends over the end of the data it should
        # return the data, repeated
        res = channel_data_cycle[slice(DATA_LENGTH, DATA_LENGTH + offset)]
        assert len(res) == offset

        testarr = np.unique(res)
        testoff = min(offset, DATA_LENGTH)

        mask = testarr == np.linspace(DATA_MIN, DATA_MAX,
                                      DATA_LENGTH)[slice(0, testoff)]
        assert mask.all()

    @given(offset=pveint)
    def test_with_earlier_slice_data_cycle(self, channel_data_cycle, offset):

        # When the mask starts before the data it should return the
        # data, repeated
        res = channel_data_cycle[slice(-offset, 0)]
        assert len(res) == offset

        testarr = np.unique(res)
        testoff = min(offset, DATA_LENGTH)

        mask = testarr == np.linspace(DATA_MIN, DATA_MAX,
                                  DATA_LENGTH)[slice(DATA_LENGTH - testoff, None)]
        assert mask.all()

    def test_throws_type_error(self, channel):

        for val in [1.0, 'str', {}, []]:
            with raises(TypeError) as err:
                channel[val]

            assert 'must be an int' in str(err.value)


class TestConstruct(object):

    def test_with_no_data(self):

        channel = Channel()

        # With no data, the internal data should be the empty
        # list and len should report zero
        assert len(channel) == 0
        assert (channel.data == np.array([])).all()

    @given(s=segment)
    def test_with_single_segment(self, s):

        frame = s[0]
        data = s[1]

        channel = Channel([s])
        cdata = channel.data

        assert len(channel) == frame + len(data)
        mask = cdata[:frame] == [data[0] for _ in range(frame)]
        assert mask.all()

        mask = cdata[frame:] == data
        assert mask.all()

    @pytest.mark.slow
    @given(intervals=lists(data(), average_size=12))
    def test_with_many_sequential_intervals(self, intervals):

        segments = []
        indices = []
        frame = 0

        for interval in intervals:
            segments.append((frame, interval))
            frame += len(interval)
            indices.append(frame)

        length = sum(map(len, intervals))

        channel = Channel(segments)

        assert len(channel) == length

        start_index = 0

        for i in range(len(indices)):
            print("Index: {}".format(indices[i]))
            print("Channel: {}".format(channel[:indices[i]]))
            print("Interval: {}".format(intervals[i]))

            mask = channel[start_index:indices[i]] == intervals[i]
            start_index = indices[i]
            assert mask.all()


    @pytest.mark.slow
    @given(intervals=lists(data(), average_size=12),
           offsets=lists(pveint, min_size=1, average_size=12))
    def test_with_many_offset_intervals(self, intervals, offsets):

        assume(len(intervals) == len(offsets))

        segments = []
        indices = []
        frame = 0

        for offset, interval in zip(offsets, intervals):
            frame += offset
            indices.append(frame)
            segments.append((frame, interval))
            frame += len(interval)
            indices.append(frame)

        length = sum(map(len, intervals)) + sum(offsets)

        channel = Channel(segments)
        assert len(channel) == length

        # TODO: Test that all the right values show up in the right
        # places, I can't think of a clean way to do this right now



import numpy as np
from math import sqrt, atan2, sin, cos


class Drawable:
    """
    A drawable consists of three things, a mask,
    a domain, and a colormap.

    The domain is a function, which given the height and
    width of the image, returns a numpy meshgrid of all the
    mathematical point in the domain each pixel represents

    The mask, is a function - which given a point in the domain
    of the drawable returns true or false indicating if that point
    is part of the drawable

    The colormap is a function - also given points in the domain
    which returns a numpy array of length 3/4 representing the
    RGB/RGBA color of that point respectively
    """

    def __init__(self, domain=None, mask=None, color=None, name=None):
        self._name = name
        self._domainfunc = domain
        self._maskfunc = np.vectorize(mask) if mask is not None else None
        self._colorfunc = np.vectorize(color, signature='(),()->(4)')\
                          if color is not None else None

    def __repr__(self):

        # The 'resolution' of the representation
        N = 8

        # We will construct a small approximation of the mask
        xdom, ydom = self.domainfunc(N, N)
        mask = self.maskfunc(xdom, ydom)

        # Little function to show us where the mask is true
        fmt = lambda t: 'XX' if t else '  '

        s = "Drawable: %s\n\n" % self.name

        # Create the 'graph' of the mask
        border = '+-' + ''.join('--' for _ in range(N)) + '-+\n'
        s += border

        for i in range(len(mask)):
            s += '| ' + ''.join(fmt(t) for t in mask[i]) + ' |\n'

        s += border + '\n'

        s += "Domain: [%.2f, %.2f] x [%.2f, %.2f]\n" % \
                (xdom[0][0], xdom[-1][-1], ydom[-1][-1], ydom[0][0])

        return s

    @property
    def name(self):
        if self._name is not None:
            return self._name
        else:
            return ''

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def domainfunc(self):
        if self._domainfunc is None:

            def default_domain(width, height):
                xs = np.linspace(-1, 1, width)
                ys = np.linspace(1, -1, height)
                return np.meshgrid(xs, ys)

            return default_domain
        else:
            return self._domainfunc

    @domainfunc.setter
    def domainfunc(self, value):
        self._domainfunc = value

    # The only purpose of this is to enable the nice decorator syntax
    # it simply passes the function to the above setter
    def domain(self, f):
        self.domainfunc = f

    @property
    def maskfunc(self):
        if self._maskfunc is None:

            def default_mask(x, y):
                return True

            return np.vectorize(default_mask)
        else:
            return self._maskfunc

    @maskfunc.setter
    def maskfunc(self, value):
        self._maskfunc = np.vectorize(value)

    # The only purpose of this is to enable the nice decorator syntax
    # it simply passes the function to the above setter
    def mask(self, f):
        self.maskfunc = f

    @property
    def colorfunc(self):
        if self._colorfunc is None:

            def default_color(x, y):
                return np.array([0, 0, 0, 255], dtype=np.uint8)

            return default_color
        else:
            return self._colorfunc

    @colorfunc.setter
    def colorfunc(self, f):
        """
        This is where we do some kung-fu introspection to try and
        optimize the code a bit better.
        """
        vcode = f.__code__
        vargs = vcode.co_varnames

        # If the color is independant of any of the arguments
        # then there is no point in evaluating the function over
        # every point in the domain, so we may as well save the result
        # now and save time later
        if len(vargs) == 0:

            # We will evaluate it and look to see if the result
            # makes sense to us
            color = f()

            if not isinstance(color, (tuple,)):
                raise TypeError('Colors must be represented by a tuple')

            # If the color is an RGB lookalike, convert it to RGBA
            if len(color) == 3:
                color = tuple([*color, 255])

            if len(color) != 4:
                raise ValueError('Colors must be in RGB (r,g,b) '
                                 'or RGBA (r,g,b,a) format!')

            self._colorfunc = color

        else:
            self._colorfunc = np.vectorize(f, signature='(),()->(4)')

    # The only purpose of this is to enable the nice decorator syntax
    # it simply passes the function to the above setter
    def colormap(self, f):
        self.colorfunc = f


def mk_domain(xmin=0, xmax=1, ymin=0, ymax=1):

    def domain(width, height):
        xs = np.linspace(xmin, xmax, width)
        ys = np.linspace(ymin, ymax, height)

        return np.meshgrid(xs, ys)

    return domain


def cartesian(X=[-1, 1], Y=[-1, 1]):
    """
    A function decorator which given the function and
    the domain it should be mapped onto, construct the
    necessary data structure so that the Image classes
    can apply it
    """

    xmin, xmax = X
    ymin, ymax = Y

    def domain(f):
        """
        This is the constructed decorator made by _cartesian_ and is
        actually where we do the mapping onto the domain
        """

        def gridfunc(width, height):

            xs = np.linspace(xmin, xmax, width)
            ys = np.linspace(ymax, ymin, height)

            return np.meshgrid(xs, ys)

        return Drawable(domain=gridfunc, mask=f, name=f.__name__)

    return domain


def extend_periodically(X=[-1, 1], Y=[-1, 1]):
    """
    A function decorator which wraps a function defined on X x Y
    and extends it to an arbitrary domain by looping it back on itself
    """

    xmin, xmax = X
    xlength = xmax - xmin
    ymin, ymax = Y
    ylength = ymax - ymin

    def wrapper(f):

        def extended(x, y):

            # I have no idea why I can't find a more efficient way
            # to do this *shrug*
            while x < xmin:
                x += xlength
            while x > xmax:
                x -= xlength
            while y < ymin:
                y += ylength
            while y > ymax:
                y -= ylength

            return f(x, y)

        extended.__name__ = f.__name__ + '_' + extended.__name__

        return extended

    return wrapper


def translate(X=[0, 0], r=0):
    """
    A function decorator which allows for easy translation of a
    domain onto the final image.
    """
    def wrapper(f):

        # We need to shift values in the opposite direction
        # to achieve the intiuitive behavior
        M = np.array([[cos(-r), -sin(-r), -X[0]],
                      [sin(-r),  cos(-r), -X[1]],
                      [0, 0, 1]])

        def transformed(x, y):

            (x, y, _) = np.dot(M, np.array([x, y, 1]))
            return f(x, y)

        transformed.__name__ = f.__name__ + '_translated'

        return transformed

    return wrapper


def reflect(Y=False, X=False):
    """
    A decorator to reflect about the x-axis, the y-axis
    or both!

    Arguments
    ---------

    X : bool, optional
        If true function is reflected across the x axis
        Default: False
    Y : bool, optional
        If true, function is reflected across the y axis
        Default: False
    """

    if X and Y:
        def wrapper(f):

            def F(x, y):

                return f(abs(x), abs(y))

            F.__name__ = f.__name__ + '_reflected'

            return F
        return wrapper

    if X:
        def wrapper(f):

            def F(x, y):

                return f(x, abs(y))

            F.__name__ = f.__name__ + '_reflected'
            return F
        return wrapper

    if Y:
        def wrapper(f):

            def F(x, y):

                return f(abs(x), y)

            F.__name__ = f.__name__ + '_reflected'
            return F
        return wrapper

    def wrapper(f):

        def F(x, y):

            return f(x, y)

        F.__name__ = f.__name__ + '_reflected'
        return F
    return wrapper


def polar():
    """
    A function decorator which constructs a wrapper function that
    additionally calculates the conversion from cartesian to polar
    coordinates and passes them into the wrapped function along with
    the original cartesian coordinates.

    This means your funtion now must be defined: f(x, y, r, t)
    """

    def domain(f):
        """
        This is the constructed decorator made by _polar_ and is
        actually where we do the mapping onto the domain
        """

        def F(x, y):

            r = sqrt(x**2 + y**2)
            t = atan2(y, x)

            return f(x, y, r, t)

        F.__name__ = f.__name__ + '_polar'

        return F

    return domain



import numpy as np

from hypothesis import given, assume
from hypothesis.strategies import floats, integers, composite
from pytest import raises
from unittest import mock

from stylo.interpolate import Sampler

# Some useful strategies
numpoints = integers(min_value=2, max_value=1000)
value = floats(min_value=-1e6, max_value=1e6)


@composite
def linear_transform(draw):
    m = draw(value)
    c = draw(value)

    return lambda x: m*x + c


class TestProperties(object):

    @given(N=numpoints)
    @mock.patch.object(Sampler, '_sample')
    def test_num_points_with_good_values(self, mock_sample, N):

        s = Sampler()

        # The sample function should already have been called
        assert mock_sample.call_count == 1

        # By default num_points should be 25
        assert s.num_points == 25

        # We should be able to change this
        s.num_points = N
        assert s.num_points == N

        # Now the _sample function should have been called again
        assert mock_sample.call_count == 2

    @given(N=integers(max_value=1, min_value=-1e6))
    @mock.patch.object(Sampler, '_sample')
    def test_num_points_with_bad_values(self, mock_sample, N):

        s = Sampler()

        # The usual defaults should have happened
        assert mock_sample.call_count == 1
        assert s.num_points == 25

        with raises(TypeError) as err:
            s.num_points = 'string'

        # We should have the correct error msg and nothing
        # should have changed
        assert 'must be an integer' in str(err.value)
        assert s.num_points == 25
        assert mock_sample.call_count == 1

        with raises(ValueError) as err:
            s.num_points = N

        assert 'be larger than 1' in str(err.value)
        assert s.num_points == 25
        assert mock_sample.call_count == 1

    @given(N=numpoints)
    def test_len_property(self, N):

        s = Sampler(num_points=N)

        # Each of num_points, len() and len(self._data)
        # should give the same value
        testarr = np.array([s.num_points, len(s), len(s._data)])
        assert (testarr == N).all()

    @given(x=value)
    @mock.patch.object(Sampler, '_sample')
    def test_f_property_with_good_values(self, mock_sample, x):

        s = Sampler()

        # The sample function should have already been called
        assert mock_sample.call_count == 1

        # The default function should be the identity
        assert s(x) == x

        # We should be able to change this
        def squared(y):
            return y * y

        s.f = squared

        # The sample function should once again have been called
        assert mock_sample.call_count == 2

        # The new function should have taken effect
        assert s(x) == x * x

    @mock.patch.object(Sampler, '_sample')
    def test_f_property_with_bad_values(self, mock_sample):

        s = Sampler()

        # The sample function should have already been called
        assert mock_sample.call_count == 1

        # The f property should be callable
        with raises(TypeError) as err:
            s.f = 2

        assert 'must be a function' in str(err.value)
        assert mock_sample.call_count == 1

    @given(a=value, b=value)
    @mock.patch.object(Sampler, '_sample')
    def test_domain_property_with_good_values(self, mock_sample, a, b):
        assume(a < b)

        s = Sampler()

        # The sample function should have been called
        assert mock_sample.call_count == 1

        # By default it should be the interval [0, 1]
        assert s.domain == [0, 1]

        # We should be able to change it
        s.domain = [a, b]
        assert s.domain == [a, b]

        # The sample function should have been called again
        assert mock_sample.call_count == 2

    @given(a=value, b=value)
    @mock.patch.object(Sampler, '_sample')
    def test_domain_property_with_bad_values(self, mock_sample, a, b):
        assume(a < b)

        s = Sampler()

        # The sample function should have been called
        assert mock_sample.call_count == 1

        # It should complain if we try and pass the wrong thing
        with raises(TypeError) as err:
            s.domain = 2.3

        assert 'must be a list in the form [a,b]' in str(err.value)
        assert s.domain == [0, 1]
        assert mock_sample.call_count == 1

        with raises(ValueError) as err:
            s.domain = [b, a]

        assert 'strictly less than' in str(err.value)
        assert s.domain == [0, 1]
        assert mock_sample.call_count == 1


class TestOperations(object):

    @given(y=value, f=linear_transform())
    def test_neg(self, y, f):

        a = Sampler(f)
        b = -a

        assert b(y) == -a(y)

    @given(y=value, f=linear_transform(), g=linear_transform())
    def test_add(self, y, f, g):

        a = Sampler(f)
        b = Sampler(g)

        # Addition of Samplers should 'do the right thing'
        c = a + b

        assert c(y) == a(y) + b(y)

    def test_add_throws_errors(self):

        a = Sampler()

        with raises(TypeError) as err:
            a + 3

        assert 'supported between Sampler objects' in str(err.value)

    @given(x=value, f=linear_transform(), g=linear_transform())
    def test_sub(self, x, f, g):

        a = Sampler(f)
        b = Sampler(g)

        # Subtraction should 'do the right thing'
        c = a - b

        assert c(x) == a(x) - b(x)

    def test_sub_throws_error(self):

        a = Sampler()

        with raises(TypeError) as err:
            a - 2

        assert 'supported between Sampler objects' in str(err.value)

    @given(x=value, f=linear_transform(), g=linear_transform())
    def test_mul(self, x, f, g):

        a = Sampler(f)
        b = Sampler(g)

        # Multiplication should 'do the right thing'
        c = a * b

        assert c(x) == a(x) * b(x)

    def test_mul_throws_errors(self):

        a = Sampler()

        with raises(TypeError) as err:
            a * 2

        assert 'supported between Sampler objects' in str(err.value)


@given(x=value, y=value)
def test_sampling_of_f(x, y):

    # Construct an interpolation of the two
    def f(t):
        return (1 - t)*x + t*y

    s = Sampler(f)

    assert s._data[0] == x
    assert s._data[-1] == y


@given(num_points=numpoints)
def test_sampler_getitem(num_points):

    s = Sampler(num_points=(num_points + 10))

    assert s[0] == 0
    assert s[-1] == 1
    assert len(s[:-10]) == num_points



import numpy as np

from math import floor, ceil, sqrt

from .coords import Drawable


class MetaSet(type):

    def __new__(cls, name, bases, attrs):

        tiles = [(key, val) for key, val in attrs.items()
                  if not key.startswith('_') and
                  isinstance(val, (Drawable,))]

        new_attrs = [(key, val) for key, val in attrs.items()
                      if (key, val) not in tiles]

        tile_list = []

        # For each tile
        for key, val in tiles:
            tile_list.append(val)

        # Add the list of tiles to the new class
        new_attrs.append(('_tiles', tile_list))

        return super().__new__(cls, name, bases, dict(new_attrs))


class TileSet(metaclass=MetaSet):
    """
    A tileset funnily enough is the counterpart to a tiled image,
    it's how you define what it is you are going to draw onto said
    image.

    NOTE: You do not use TileSet directly! Instead create a class
    which inherits from this one!
    """

    def __init__(self, FPS=25):
        self._FPS = FPS

    def __getitem__(self, key):

        if isinstance(key, (int,)):
            frame = key
            time = key / self._FPS

        if isinstance(key, (float,)):
            frame = floor(key * self._FPS)
            time = key

        layout = self.layout(frame, time)

        return (self._tiles, layout)

    def layout(self, frame, time):
        """
        This implements the default layout, which simply
        displays all available tiles in a grid
        """

        num_tiles = len(self._tiles)

        # Find out what the dimension of the grid will be
        N = ceil(sqrt(num_tiles))

        # Make that many Nones
        nones = [None for _ in range(N*N)]

        # Replace with the appropriate numbers
        nones[:num_tiles] = range(num_tiles)

        # Finally convert to numpy array and
        # make it 2D
        grid = np.array(nones, dtype='O')
        return grid.reshape(N, N)



import multiprocessing as mp


def animate(f, start=0, stop=10, fps=25, frames=None):
    """
    Given a function `f`, with only the frame number as a
    parameter, evaluate it in parallel to quickly construct
    all frames.

    You can either specify start and stop times (in seconds)
    and the fps or just specify the number of total frames.
    """

    pool = mp.Pool()

    if frames is not None:
        num_frames = frames
    else:
        num_frames = (stop - start) * fps
        pool.map(f, range((stop - start) * fps))

    print('Animating...')
    pool.map(f, range(num_frames))

    pool.close()



import numpy as np
import matplotlib.pyplot as plt


from itertools import cycle
from math import floor


class Channel:
    """
    A Channel is a single 'track' of Sampler objects and the
    like and is resonsible for all time conversions etc. everything
    it takes to construct, index and manage all the data for a single
    animation track.
    """

    def __init__(self, segments=[], FPS=25, name='', cycle=False):
        self._segments = segments
        self._FPS = FPS
        self._name = name
        self._cycle = cycle
        self._data = []

        self._construct()

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        # Although it is enough to define __getitem__ to make an object
        # iterable we have extended it to essentially make this object
        # infinitely iterable. This means any for loop will never terminate
        # so we have to define __iter__ to return the actual data we have so
        # we hopefully avoid that problem
        return iter(self._data)

    def _gen_fake_data(self, index):

        if isinstance(index, (int,)):
            return 0.0

        if isinstance(index, (slice,)):
            return np.array([0.0 for _ in range(index.start, index.stop)])

    def _get_single(self, index):

        if self._cycle:
            return self._data[index % len(self._data)]

        if index > len(self._data) - 1:
            return self._data[-1]

        if index < 0:
            return self._data[0]

        return self._data[index]

    def _get_slice(self, index):

        # Unpack the slice
        start = 0 if index.start is None else index.start
        stop = len(self._data) if index.stop is None else index.stop
        data_length = len(self._data)

        # Some conditions
        must_extend = stop > data_length
        must_prepend = start < 0

        # Handle the simplest case
        if not must_prepend and not must_extend:
            return self._data[index]

        # Oh dear, looks like we need to do some fancy stuff...
        # To start with we will extract the inner region of the
        # data and then decide on which directions it might need
        # extending
        inner_start = max(0, start)
        inner_end = min(stop, data_length)
        inner_data = self._data[slice(inner_start, inner_end)]

        # Assume we don't need to extend anything
        extension = []
        prepension = []

        # Do we have to extend it?
        if must_extend:
            # By how much?
            length = stop - inner_end

            # With what?
            if self._cycle:

                data = cycle(self._data)
                extension = [next(data) for _ in range(length)]

            else:
                extension = [self._data[-1] for _ in range(length)]

        # Do we have to prepend it?
        if must_prepend:
            # By how much?
            length = abs(start)

            # With what?
            if self._cycle:

                data = cycle(reversed(self._data))
                prepension = list(
                                reversed([next(data) for _ in range(length)]))

            else:
                prepension = [self._data[0] for _ in range(length)]

        # Finally return the 'new data'
        return np.append(np.append(prepension, inner_data), extension)

    def __getitem__(self, index):

        if not isinstance(index, (int, slice)):
            raise TypeError('indices must be an integer or slices!')

        if self._segments == []:
            return self._gen_fake_data(index)

        if isinstance(index, (int,)):
            return self._get_single(index)

        if isinstance(index, (slice,)):
            return self._get_slice(index)

    @property
    def data(self):
        return self._data

    @property
    def cycle(self):
        return self._cycle

    @cycle.setter
    def cycle(self, value):

        if not isinstance(value, (bool,)):
            raise TypeError('Cycle property must be a bool!')

        self._cycle = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):

        if not isinstance(value, (str,)):
            raise TypeError('Name property must be a string!')

        self._name = value

    @property
    def FPS(self):
        return self._FPS

    @FPS.setter
    def FPS(self, value):

        if not isinstance(value, (int,)):
            raise TypeError('FPS property must be an integer!')

        if value <= 0:
            raise ValueError('FPS property must be larger than 0!')

        self._FPS = value
        self._construct()

    @property
    def segments(self):
        return self._get_sorted_segments()

    @segments.setter
    def segments(self, value):

        # Before we do anything make sure the segments are
        # given in the right format
        if not isinstance(value, (list,)):
            raise TypeError('Segments property must be a list!')

        for item in value:

            if not isinstance(item, (tuple,)) or len(item) != 2:
                raise TypeError('Each segment must be a tuple of '
                                'length 2!')

            # Unpack the tuple
            frame = item[0]

            if not isinstance(frame, (int,)):
                raise TypeError('The frame number must be an integer!')

            if frame < 0:
                raise ValueError('The frame number cannot be negative!')

            data = item[1]

            try:
                (v for v in data)
            except:
                raise TypeError('The segment data must be iterable!')

        # Only now do we accept it and do our thing
        self._segments = value
        self._construct()

    def _get_sorted_segments(self):
        return sorted(self._segments, key=lambda x: x[0])

    def _construct(self):
        # This function takes the segments and combines them
        # into a single continuous array of data.
        #
        # To make things simple, we assume that the data starts
        # at frame 0 and if a segment starts before the previous
        # has finished then we shift it so that it starts immediately
        # afterwards

        # Get the segments and setup
        segments = self._get_sorted_segments()
        array = []
        current_time = 0

        for segment in segments:

            # Unpack the segment
            frame, data = segment

            if frame > current_time:

                # Pad the data until its time to start with the
                # previous value
                value = data[0] if len(array) == 0 else array[-1]
                array = np.append(array,
                                  [value for _ in range(frame - current_time)])

            # Add on the data
            array = np.append(array, data)
            current_time = len(array)

        # Set the data
        self._data = np.array(array)


class Driver:
    """
    A Driver can be used to control a Puppet, it has one or
    more channels that represent a single 'string' on a given
    puppet. It is responsible for ensuring consistency across
    each of the tracks.
    """

    def __init__(self, FPS=25):
        self._channels = {}
        self._FPS = FPS

    def __len__(self):

        lengths = []

        for name in self._channels:
            channel = self.__getattribute__(name)
            lengths.append(len(channel))

        return max(lengths)

    def _to_frame_number(self, realtime):

        if isinstance(realtime, (int,)):
            return realtime

        return floor(realtime * self.FPS)

    def _get_from_str(self, index):

        if index not in self._channels:
            raise IndexError('No such Channel: {}'.format(index))

        return self.__getattribute__(index)

    def __getitem__(self, index):

        if isinstance(index, (str,)):
            return self._get_from_str(index)

    def add_channel(self, name, channel=None, offset=0.0, *args, **kwargs):
        """
        Add a Channel to the Driver object.

        Either you can provide a name and the relevant __init__ arguments
        to this method and a new Channel instance will be created and added
        under the given name.

        Or you can provide a name and an existing Channel instance which will
        be added to the Driver under the given name

        Parameters:
        -----------

        name: str
            The name you want to give to the Channel
        offset: float, optional
            The time you want to delay the start of the channel by, in seconds
            Default: 0.0
        channel : Channel, optional
            The instance of the Channel you want to add.
            Default: None

        See the documentation for the Channel's __init__ for more information
        on what can be passed to this function
        """

        if not isinstance(name, (str,)):
            raise TypeError('Channel names must be a string!')

        if name == '':
            raise ValueError('Channel names cannot be the empty string')

        if name in self._channels:
            raise RuntimeError('A Channel with the name "{}" '.format(name)
                               + 'already exists!')

        if channel is not None and not isinstance(channel, (Channel,)):
            raise TypeError('Expected Channel instance, got ' +
                            '{} instead'.format(type(channel)))

        # Did the user provide an exisiting channel?
        if channel is not None:
            # Enforce the new name on the channel
            channel.name = name
            self.__setattr__(name, channel)

        else:
            # Otherwise add new Channel object, accessible as an attribute
            # of the same name
            self.__setattr__(name, Channel(name=name, *args, **kwargs))

        # Don't forget to add the name to the internal record
        self._channels[name] = offset

    def del_channel(self, name):
        """
        Delete a Channel from the Driver instance.

        Given a name representing the channel you want deleted
        the method removes the channel instance from the driver.

        Parameters:
        -----------

        name: str
            The name of the channel you want deleted
        """

        if not isinstance(name, (str,)):
            raise TypeError('Channel names must be referenced by a string')

        if name not in self._channels:
            raise RuntimeError('Channel name "{}" '.format(name) +
                               'does not exist!')

        # Delete the Channel instance
        self.__delattr__(name)

        # Remove the entry from the internal record
        del self._channels[name]

    @property
    def channels(self):
        return list(self._channels.keys())

    @property
    def FPS(self):
        return self._FPS

    @FPS.setter
    def FPS(self, value):

        if not isinstance(value, (int,)):
            raise TypeError('FPS property must be an integer!')

        if value < 1:
            raise ValueError('FPS property must be larger than zero!')

        self._FPS = value


class Sampler:

    def __init__(self, f=None, num_points=25, name=None, domain=[0, 1]):

        if f is not None and not callable(f):
            raise TypeError('f must be a function!')

        self._f = f
        self._num_points = num_points
        self._name = name
        self._domain = domain
        self._sample()

    def __getitem__(self, index):
        return self._data[index]

    def __call__(self, x):

        # Default lambda x: x
        if self._f is None:
            return x
        else:
            return self._f(x)

    def __repr__(self):
        s = '%s\n' % self.name
        s += 'Num Points: %s' % self.num_points
        return s

    def __len__(self):
        return len(self._data)

    def __neg__(self):
        return Sampler(lambda x: -self.f(x))

    def __add__(self, other):

        if not isinstance(other, (Sampler,)):
            raise TypeError('Addition is only supported '
                            'between Sampler objects!')

        return Sampler(lambda x: self.f(x) + other.f(x))

    def __sub__(self, other):

        if not isinstance(other, (Sampler,)):
            raise TypeError('Subtraction is only supported '
                            'between Sampler objects!')

        return Sampler(lambda x: self.f(x) - other.f(x))

    def __mul__(self, other):

        if not isinstance(other, (Sampler,)):
            raise TypeError('Multiplication is only supported '
                            'between Sampler objects!')

        return Sampler(lambda x: self.f(x) * other.f(x))

    def _sample(self):
        a, b = self.domain
        points = np.linspace(a, b, self._num_points)

        if self._f is None:
            self._data = points
        else:
            self._data = np.array([self._f(t) for t in points])

    @property
    def domain(self):
        return self._domain

    @domain.setter
    def domain(self, value):

        if not isinstance(value, (list,)):
            raise TypeError('Domain property must be a list in the form [a,b]')

        a, b = value

        if a >= b:
            raise ValueError('The start of the domain must be strictly less '
                             'than the end!')

        self._domain = value
        self._sample()

    @property
    def f(self):
        if self._f is None:
            return lambda t: t
        else:
            return self._f

    @f.setter
    def f(self, value):

        if not callable(value):
            raise TypeError('f property must be a function!')

        # I'd love to use __code__.co_argcount to make sure we are given
        # a function in a single argument. However this breaks for
        # builtin functions such as math.cos

        self._f = value
        self._sample()

    @property
    def num_points(self):
        return self._num_points

    @num_points.setter
    def num_points(self, value):

        if not isinstance(value, (int,)):
            raise TypeError('num_points must be an integer!')

        if value <= 1:
            raise ValueError('num_points must be larger than 1!')

        self._num_points = value
        self._sample()

    @property
    def name(self):

        if self._name is None:
            return 'Sampled function:'
        else:
            return self._name

    @name.setter
    def name(self, value):

        if not isinstance(value, (str,)):
            raise TypeError('property name must be a string!')

        self._name = value

    def show(self):
        """
        Returns a matplotlib figure of both the function and the sampled points
        """
        a, b = self.domain
        points = np.linspace(a, b, self._num_points)
        interval = np.linspace(a, b, 512)
        fs = [self._f(x) for x in interval]

        plt.plot(interval, fs, 'k')
        return plt.scatter(points, self._data, c='k')


def sampled(num_points=25):

    def sampler(f):

        name = f.__name__

        return Sampler(f, num_points=num_points, name=name)

    return sampler


def linear(x0, x1, num_points=25):
    """
    Linearly interpolate between x0 and x1 between times 0 and 1
    """
    def f(t):
        return (1 - t)*x0 + t*x1

    name = 'Linear Interpolation\nFrom:\t%s\nTo:\t%s\n' % (str(x0), str(x1))

    return Sampler(f, num_points=num_points, name=name)


def quadratic_ease_in(x0, x1, a=1):
    """
    Quadratic interpolation between x0 and x1
    which eases in. tuning parameter a gives control over
    the trajectory of the interpolation
    """

    # Make sure that the parameter is always positive
    a = abs(a)

    def f(t):
        return a*t**2 + t*(x1 - x0 - a) + x0

    return Sampler(f)


def quadratic_ease_out(x0, x1, a=-1):
    """
    Quadratic interpolation between x0 and x1
    which eases out tuning parameter a gives control over
    the trajectory of the interpolation
    """

    # Make sure that the parameter is always negative
    a = -abs(a)

    def f(t):
        return a*t**2 + t*(x1 - x0 - a) + x0

    return Sampler(f)



from .coords import cartesian, extend_periodically, translate, reflect,\
                    polar, mk_domain
from .color import hex_to_rgb
from .image import Image, LayeredImage, TiledImage
from .objects import TileSet
from .prims import between, circle, ellipse, rectangle, square
from .time import animate

from .version import __version__



import pytest
import numpy as np

from stylo.interpolate import Driver, Channel

from pytest import raises
from hypothesis import given, assume
from hypothesis.strategies import text, lists, integers,\
                                  composite, floats


# Some useful strategies
names = text(min_size=1, average_size=12)
pveint = integers(min_value=1, max_value=1024)
real = floats(min_value=-1e6, max_value=1e6)


@composite
def data(draw):
    length = draw(pveint)
    start = draw(real)
    stop = draw(real)

    return (0, np.linspace(start, stop, length))


class TestProperties(object):

    @given(c_names=lists(names, average_size=4))
    def test_channels_property(self, c_names):
        assume(len(set(c_names)) == len(c_names))

        driver = Driver()

        for name in c_names:
            driver.add_channel(name)

        # This should simply return a list of the available
        # channels
        assert set(driver.channels) == set(c_names)

        # This attribute should be read only
        with raises(AttributeError) as err:
            driver.channels = 2

        assert 'can\'t set attribute' in str(err.value)

    @given(fps=pveint)
    def test_fps_property_with_good_values(self, fps):

        driver = Driver()

        # By default fps should be 25
        assert driver.FPS == 25

        # We should be able to change this
        driver.FPS = fps
        assert driver.FPS == fps

    @given(fps=pveint)
    def test_fps_property_with_bad_values(self, fps):

        driver = Driver()

        with raises(TypeError) as err:
            driver.FPS = 'a str'

        assert 'must be an integer' in str(err.value)

        with raises(ValueError) as err:
            driver.FPS = -fps

        assert 'must be larger than zero' in str(err.value)

    @pytest.mark.slow
    @given(segments=lists(data(), min_size=1, average_size=4),
           names=lists(names, min_size=1, average_size=4))
    def test_len_property(self, names, segments):

        assume(len(names) == len(set(names)))
        assume(len(segments) == len(names))
        lengths = []

        driver = Driver()

        for name, data in zip(names, segments):
            driver.add_channel(name, segments=[data])
            lengths.append(len(data[1]))

        # When we ask for the length of the driver, it should be
        # be equal to the longest of its component channels
        assert len(driver) == max(lengths)


class TestChannelManagement(object):

    @given(name=names)
    def test_adding_new_channel(self, name):

        driver = Driver()
        driver.add_channel(name, FPS=50, cycle=True)

        # The name should have been added into the internal record
        # of channels
        assert name in driver._channels

        # We should now be able to access the Channel object
        # directly
        assert isinstance(driver.__getattribute__(name), (Channel,))

        # It should also have respected its own __init__ arguments
        assert driver.__getattribute__(name).FPS == 50
        assert driver.__getattribute__(name).cycle

    @given(name=names)
    def test_adding_existing_channel(self, name):

        channel = Channel()
        driver = Driver()

        driver.add_channel(name, channel=channel)
        assert name in driver._channels
        assert isinstance(driver.__getattribute__(name), (Channel,))

        # By default the Channel should be added with a zero offset.
        assert driver._channels[name] == 0

    @given(name=names)
    def test_add_channel_throws_errors(self, name):

        driver = Driver()

        # Name should only be a str
        with raises(TypeError) as err:
            driver.add_channel(1)

        assert 'must be a string' in str(err.value)

        # And the string cannot be empty
        with raises(ValueError) as err:
            driver.add_channel('')

        assert 'cannot be the empty string' in str(err.value)

        driver.add_channel(name)

        # Adding a new channel with an exisiting name should throw an
        # error
        with raises(RuntimeError) as err:
            driver.add_channel(name)

        assert 'already exists' in str(err.value)

        with raises(TypeError) as err:
            driver.add_channel('new name', 3)

        assert 'Expected Channel instance' in str(err.value)

    @given(name=names)
    def test_deleting_channel(self, name):

        driver = Driver()
        driver.add_channel(name)

        assert name in driver._channels
        assert isinstance(driver.__getattribute__(name), (Channel,))

        driver.del_channel(name)

        assert name not in driver._channels

        with raises(AttributeError):
            driver.__getattribute__(name)

    @given(name=names)
    def test_del_channel_throws_errors(self, name):

        driver = Driver()

        with raises(TypeError) as err:
            driver.del_channel(2)

        assert 'by a string' in str(err.value)

        with raises(RuntimeError) as err:
            driver.del_channel(name)

        assert 'does not exist' in str(err.value)


class TestGetItem(object):

    @given(name=names)
    def test_get_channel_by_name(self, name):

        driver = Driver()
        driver.add_channel(name)

        channel = driver[name]

        assert isinstance(channel, (Channel,))



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


exec(open('stylo/version.py', 'r').read())

setup(name='stylo',
      version=__version__,
      description='Using a blend of Python and Maths for the '
      'creation of images',
      long_description=readme(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: Multimedia :: Graphics'
      ],
      author='Alex, Carney',
      author_email='alcarneyme@gmail.com',
      license='MIT',
      packages=['stylo'],
      install_requires=[
          'matplotlib',
          'numpy',
          'Pillow',
          'scipy'
      ],
      setup_requires=['pytest-runner'],
      test_suite='tests',
      tests_require=['pytest',
                     'hypothesis'],
      python_requires='>=3.0',
      include_package_data=True,
      zip_safe=False)



from math import floor


def ellipse(x0, y0, a, b, r, pt=0.2, fill=False):
    """
    Mathematically we can define an ellipse to be the set
    of points :math:`(x, y)` which satisfy:

    .. math::

        \\frac{(x - x_0)^2}{a^2} + \\frac{(y - y_0)^2}{b^2} = r^2

    where:

    - :math:`(x_0, y_0)` is the center of the ellipse
    - :math:`a` is known as the semi major axis, larger values make the
      ellipse more elongated in the :math:`x`-direction
    - :math:`b` is known as the semi minor axis, larger values make the
      ellipse more elongated in the :math:`y`-direction
    - :math:`r` is the "radius" of the ellipse and controls the overall
      size of the ellipse

    This function will return another function that when given a point
    :code:`(x,y)` that will return :code:`True` if the point is in the
    ellipse.

    Parameters
    ----------

    x0 : float
        Represents the x coordinate of the ellipse's center
    y0 : float
        Represents the y coordinate of the ellipse's center
    a : float
        Represents the semi major axis of the ellipse
    b : float
        Represents the semi minor axis of the ellipse
    r : float
        Represents the "radius" of the ellipse
    pt : float, optional
        Represents the thickness of the line of the ellipse.
        Default: 0.2
    fill : bool, optional
        Fill the ellipse rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt

    Returns
    -------

    function:
        A function in 2 arguments :code:`(x,y)` which returns :code:`True`
        if that point is in the ellipse defined by the above parameters
    """

    lhs = lambda x, y: ((x - x0)**2)/a**2 + ((y - y0)**2)/b**2

    if fill:

        def ell(x, y):

            if lhs(x, y) <= r**2:
                return True

            return False

        return ell

    else:

        def ipse(x, y):

            val = lhs(x, y)

            if val <= (r + pt)**2 and val >= (r - pt)**2:
                return True

            return False
        return ipse


def circle(x0, y0, r, *args, **kwargs):
    """
    Mathematically a circle can be defined as the set of all
    points :math:`(x, y)` that satisfy

    .. math::

        (x - x_0)^2 + (y - y_0)^2 = r^2

    This function returns another function which when given
    a point :code:`(x, y)` will return :code:`True` if that
    point is in the circle

    Parameters
    ----------

    x0 : float
        This is the x coordinate of the circle's center
    y0 : float
        This is the y coordinate of the circle's center
    r : float
        This represents the radius of the ellipse
    pt : float, optional
        Represents the thickness of the lines of the circle.
        Default: 0.2
    fill : bool, optional
        Fill the circle rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt

    Returns
    -------

    function:
        A function in 2 arguments :code:`(x, y)` that returns
        :code:`True` if that point is in the circle defined by the
        above parameters
    """

    return ellipse(x0, y0, 1, 1, r, *args, **kwargs)


def between(lower, value, upper):
    """
    A simple function which provides a shorthand
    for checking if a given value is between some lower
    and upper bound

    Parameters
    ----------
    lower : float
        The lower bound to check
    value : float
        The value you want checked
    upper: float
        The upper bound to check

    Returns
    -------
    bool
        :code:`True` if :code:`lower <= value` and
        :code:`value <= upper`. :code:`False` otherwise
    """
    return lower <= value and value <= upper


def rectangle(x0, y0, width, height, pt=0.2, fill=False):
    """
    It's quite simple to define a rectangle, simply pick a
    point :math:`(x_0,y_0)` that you want to be the center
    and then two numbers which will represent the width and
    height of the rectangle.

    Parameters
    ----------

    x0 : float
        Represents the x-coordinate of the rectangle's center
    y0 : float
        Represents the y-coordinate of the rectangle's center
    width : float
        Represents the width of the rectangle
    height : float
        Represents the height of the rectangle
    pt : float, optional
        Represents the thickness of the lines of the rectangle.
        Default: 0.2
    fill : bool, optional
        Fill the rectangle rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt

    Returns
    -------

    function
        A function in 2 arguments :code:`(x, y)` that returns :code:`True`
        if the point is in the rectangle defined by the above parameters
    """
    left = x0 - (width / 2)
    right = x0 + (width / 2)
    top = y0 + (height / 2)
    bottom = y0 - (height / 2)

    def rect(x, y):

        if x >= left and x <= right and\
           y >= bottom and y <= top:
            return True

        return False

    if fill:
        return rect

    def small(x, y):

        if x >= left + pt and x <= right - pt and\
           y >= bottom + pt and y <= top - pt:
            return True

        return False

    def test(x, y):

        if rect(x, y) and not small(x, y):
            return True

        return False

    return test


def square(x0, y0, size, *args, **kwargs):
    """
    It's quite simple to define a square, simply pick a
    point :math:`(x_0,y_0)` that you want to be the center
    and then a number which will represent the size of the
    square.

    Parameters
    ----------

    x0 : float
        Represents the x-coordinate of the square's center
    y0 : float
        Represents the y-coordinate of the square's center
    size : float
        Represents the size of the square
    pt : float, optional
        Represents the thickness of the lines of the square.
        Default: 0.2
    fill : bool, optional
        Fill the square rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt

    Returns
    -------

    function
        A function in 2 arguments :code:`(x, y)` that returns :code:`True`
        if the point is in the square defined by the parameters above
    """
    return rectangle(x0, y0, size, size, *args, **kwargs)
